require 'singleton'

require 'bf/state/clientstate'
require 'bf/state/gamestate'
require 'bf/client/log'
require 'bf/client/recv'
require 'bf/client/send'
require 'bf/net/session'
require 'bf/games/include'

class GameClient
  include Singleton
  attr_accessor :playername, :connected, :session, :serveruri
  attr_accessor :ui

  TIMEOUT = 20

  # need to set the uri and ui
  def initialize
    @log = Logger['bf::client']
    @log.info "Setting up client"
    @mutex = Mutex.new
  end

  def connected?; not @session.nil? and @session.connected? end

  def connect
    if connected?
      @ui.write_cmsg  "Already connected."
      return
    elsif @serveruri.nil?
      @ui.write_cmsg "No server selected."
      return
    end
    # ok, let's connect to server
    begin
      create_sockets
      listen
      time = 0
      # wait a bit until connected, or timeout
      while time < TIMEOUT and not @session.connected?
        sleep 0.5 
        time += sleep
      end
      # register with server
      netsend :recv_register_player, @playername
    rescue Errno::ECONNREFUSED
      @ui.write_cmsg \
      "Connection to #{@serveruri} refused. Is it the right address?"
    rescue Exception => e
      @ui.write_cmsg "Error: unable to connect to #{@serveruri}"
      @log.error "Unable to establish connection to #{@serveruri}"
      @log.debug e
      @ui.server_shutdown
    end
  end

  def disconnect
    return if not connected?
    @log.info "Disconnecting client"
    netsend :recv_client_disconnect
    @session.close
  end

  def create_sockets
    if @serveruri =~ %r{^bf://(.*):(.*)}
      host = ($1 == "" ? nil : $1)
      listenport = $2.to_i
      writeport = listenport + 1
    else
      raise ArgumentError, "Invalid uri", caller
    end
    listenio = TCPSocket.new(host, listenport)
    @session = Session.new(listenio)
  end

  # method that listens to the server
  def listen
    Thread.new {
      if @session.connected?
        begin
          while (msg = @session.session_get)
            if msg[1].nil? then self.send(msg[0])
            else self.send(*msg)
            end
          end
        rescue EOFError
          #@log.warn "Server disconnected abruptily"
          recv_server_disconnect "server disconnect"
        rescue Exception => e
          @log.debug {e}
          disconnect
        end
      end
    }
    @log.info "created listener" 
  end

  def netsend(method, obj=nil)
    trap_neterr { @session.session_send(method, obj) }
  end

  def trap_neterr
    if connected?
      @mutex.synchronize {
        begin yield
        rescue Exception => e
          @log.error "Bad session trapped."
          @log.debug {e}
          disconnect "bad session"
          @mutex.unlock
        end
      }
    else @ui.write_cmsg "Not connected to a server."
    end
  end

end
