class Board
  attr_reader :name, :author, :rows, :cols, :squares

  def initialize(*args)
    @name, @author, @rows, @cols, @squares = args
    init_deps
  end

  # gets the coords of the next square in a given direction
  def next_square_pos(dir, row, col)
    y, x = row, col
    case dir
    when N then y -= 1
    when W then x -= 1
    when S then y += 1
    when E then x += 1
    end
    return wrap(y, x) if wrap? and out_of_bounds? y, x
    return [y, x]
  end

  # given a point, get the square
  def square_at_coord(x, y)
    size = $game.theme['square_size']
    square_at( y / size, x / size )
  end

  # returns the square at row, col
  # if it is out of bounds, return the wrapped square
  # otherwise, return the null square
  def square_at(row, col)
    return @squares[row * @cols + col] unless out_of_bounds?(row, col)
    return square_at(*wrap(row, col)) if wrap?
    NoSquare.new
  end
  def [](row, col)
    square_at(row,col)
  end

  def next_square(dir, square)
    square_at(*next_square_pos(dir, square.row, square.col))
  end
  def next_square2(dir, row, col)
    square_at(*next_square_pos(dir, row, col))
  end

  # implement later
  def wrap?; false end
  # wraps an out of bounds coordinate
  def wrap(row, col)
    y, x = row, col
    y = @rows - 1 if row < 0
    x = @cols - 1 if col < 0
    y = 0 if row > @rows - 1
    x = 0 if col > @cols - 1
    return y, x
  end


  def [](row, col); square_at(row, col) end

  def each_square; @square.each{|s| yield s} end

  # all 8 squares around a coord, NoSquare if none present
  def surrounding_squares(square)
    row, col = square.row, square.col
    nw, sw, se, ne = [NoSquare.new]*4
    n = next_square(N, square)
    unless n.kind_of?( NoSquare )
      nw = next_square(W, n)
      ne = next_square(E, n)
    end
    w = next_square(W, square)
    s = next_square(S, square)
    unless s.kind_of?( NoSquare )
      sw = next_square(W, s)
      se = next_square(E, s)
    end
    e = next_square(E, square)
    return [n,w,s,e], [nw,sw,se,ne]
  end

  # only the four tha share edges
  def adjacent_squares(s)
    return [next_square(N,s),next_square(W,s),next_square(S,s),next_square(E,s)]
  end

  # is there a wall in the way?
  # check both this square and the next square
  def wall_in_way?(dir, square)
    return true if square.wall_at?(dir)
    next_square(dir, square).wall_at?(dir.rear)
  end

  def out_of_bounds?(row, col)
    return true if row.nil? or col.nil?
    return true if row < 0 || col < 0
    return true if row >= @rows || col >= @cols
    false
  end

  # for each square, init data that depend on other squares
  def init_deps
    @squares.each {|s| s.init_deps(self) }
  end

  # whenever a permanent fixture that can block lasers is changed, this
  # should be called to rebuild the laser deps
  def rebuild_lasers
    sources = []
    for s in @squares
      l = s.get(Lasers)
      next if l.nil?
      if l.is_source?
        l.reinitialize
        sources << s
      else
        s.elements.delete(l)
      end
    end
    sources.each{ |s| s.init_deps(self) }
  end

  #### robots stuff

  def squares_with_robots
    s = $game.game_robots.map{ |r| square_at(r.row, r.col) }
    s.compact!
    s.uniq!
    s
  end

  # finds a robot in robot's los
  def robot_in_los(robot, facing)
    for r in $game.game_robots
      next if r == robot
      case facing
      when N, S 
        if r.col == robot.col
          return r if has_los?(facing, self[robot.row,robot.col], self[r.row,r.col])
        end
      when W, E
        if r.row == robot.row
          return r if has_los?(facing, self[robot.row,robot.col], self[r.row,r.col])
        end
      end
    end
    nil
  end

  # returns true if start square has los to target square
  def has_los?(facing, square, target)
    while true
      square = next_square(facing, square)
      return false if out_of_bounds?(square.row, square.col)
      if square == target
        return false if wall_in_way?(facing, square)
        return true
      end
    end
  end

  def robots_at(row, col)
    ret = []
    for robot in $game.game_robots
      ret << r if r.row == row and r.col == col
    end
    ret
  end

end
