# builds the internal board from an XML file

require 'rexml/document'

def decode_facings(array)
  array.map! {|e| Facing.new(e)}
  array
end

# some convenience methods
module REXML
  class Element
    # get elmt's text whether an attribute or leaf
    def value_of(elmt)
      val = attributes[elmt]
      if val.nil?
        sub = elements[elmt]
        val = sub.text unless sub.nil?
      end
      val
    end
    # split comma-delimited text into an array
    def split_text(name)
      txt = value_of(name)
      return [] if txt.nil? or txt.strip.empty?
      txt.split(/\s*,\s*/)
    end
  end
end

class DefinitionError < Exception
end

class BoardFactory
  include REXML

  def self.create_from_file(filename)
    file = File.new(filename)
    xml = file.readlines.join
    return create_from_string(xml)
  end

  def self.create_from_string(xml)
    return xml, decode_xml(Document.new(xml))
  end

  def self.decode_xml(doc)
    boarddef = doc.elements['/board']
    if boarddef.nil?
      raise ParseException, "No board defined!", caller[1..-1]
    end
    return decode_board(boarddef)
  end

  def self.decode_board(doc)
    # board info
    name = doc.value_of('name')
    author = doc.value_of('author')
    rows = doc.value_of('rows').to_i
    cols = doc.value_of('cols').to_i

    # populate the board with squares
    squares = []
    square_elmts = XPath.match(doc, '/board/square')
    for row in 0...rows
      for col in 0...cols
        squares << SquareFactory::decode_square(row, col, square_elmts[row*cols+col])
      end
    end

    return Board.new(name, author, rows, cols, squares)
  end
end

class SquareFactory
  def self.decode_square(row, col, e)
    # walls
    faces = e.split_text('walls')
    walls = Walls.new(*decode_facings(faces))
    floor = decode_floor(e)
    # other elements
    elements = decode_elements(e)
    # build the square
    hash = {
      :row=>row, :col=>col, 
      :floor=>floor, :elements=>elements, :walls=>walls
    }
    return Square.new(hash)
  end

  # basic for now
  def self.decode_floor(e)
    ret = PlainFloor.new
    sub_e = nil
    if (sub_e = e.elements['pit'])
      faces = sub_e.split_text('continue')
      pit_cont = Sides.new(*decode_facings(faces))
      ret = Pit.new
    end
    if (sub_e = e.elements['conveyor'])
      ret = decode_conveyor(sub_e)
    end
    ret
  end

  def self.decode_elements(xml)
    ret = []
    if (e = xml.elements['laser'])
      ret << decode_laser(e)
    end
    ret
  end
end
