# standalone game, where multiple players play on one computer
# without a network connection.

class StandaloneGame
  def initialize(ui)
    @log = Logger['bf::client']
    @log.info 'Starting StandaloneGame'

    @ui = ui

    # robots will hold the players keyed to their names
    $game = GameState.new
  end

  # the game should be initialized with the number of players, the game
  # type and the board, all from the ui
  def create_game(playernames)
    i = 1 # this ddserves as the clientid
    for name in playernames
      player = Robot.new(i, name)
      $game.robots[name] = player
      i += 1
    end
    $game.game_robots = $game.robots.values
    start_game
  end
  
  # entry point when players are ready_to_play
  def start_game
    @log.info "Game is #{$game.gametype.class}"
    @log.debug {$game.gametype}
    $game.game_running = true
    $game.deck.shuffle
    #assign_robots
    $game.gametype.adjust($game)
    $game.game_robots.each {|r| r.virtual = true}
    do_turn
  end

  def recv_play_hand(robotid, hand)
    robot = $game.robots[robotid]
    robot.program_complete = true
    begin
      robot.handle_play(hand)
    rescue CheaterError => ce
      @log.warn "Cheater detected: #{ce.message}"
      kick_robot(name, "for cheating")
      return
    end
    # wait until all robots have finished programming
    for r in $game.game_robots
      return if not r.program_complete
    end
    begin
      Thread.new { compute_turn }.abort_on_exception = true
    rescue Exception => e
      @log.fatal e
      raise e
    end
  end

  # entry point at begining of each turn
  def do_turn
    $game.game_robots.each {|p| p.program_complete = false}
    do_programming_phase
  end

  def do_programming_phase
    $game.deck.shuffle
    deal_hands
  end

  def deal_hands
    $game.game_robots.each do |robot|
      $game.deck.deal_hand(robot)
    end
    @ui.recv_hands # returns control to ui
  end

  # once veryone has submitted their hands
  def compute_turn
    $game.turn += 1
    $game.turn_history = []
    @log.debug "computing turn #{$game.turn}"
    $game.record(Action::PAUSE)
    (1..5).each {|num| compute_phase(num) }
    end_of_turn
    prep_next_turn
    $game.game_history << $game.turn_history.clone
    end_game_by_objective if $game.gametype.game_over?
    @ui.draw_turn
    do_turn # goto do_turn
  end

  # TODO: animate all robots simultaneously after each phase
  def compute_phase(phase)
    @log.debug "computing phase #{phase}"
    @log.debug "  - robot movement"
    $game.phase = phase
    robots_move(phase)
    board_components_act(phase)
    resolve_laser_fire
    end_of_phase
  end

  def robots_move(phase)
    $game.live_robots.sort.reverse.each {|r| r.execute_phase(phase)}
  end

  def board_components_act(phase)
    for priority in 1..6
      action_squares = $game.board.squares_with_robots
      for s in action_squares
        s.immediate_effects # need to do this in case robot is pushed on pit 
        s.execute(phase, priority)
      end
    end
    $game.board.squares_with_robots.each {|s| s.immediate_effects}
  end

  # robots only TODO: debug case where fired > 0 for one robot
  def resolve_laser_fire
    fired = $game.live_robots.collect{|r| r.fire_laser}
    fired.delete(false)
    $game.record(Action::LASERFIRE) if fired.size > 0
    $game.live_robots.each {|r| r.die_if_dead!}
  end

  # kind of redundant
  def end_of_phase
    return
    $game.board.squares_with_robots.each {|t| t.end_of_phase}
  end

  def end_of_turn 
    @log.debug "computing end of turn"
    $game.board.squares_with_robots.each {|t| t.end_of_turn}
    toggle_virtual
  end

  def toggle_virtual
    $game.virtual_robots.each {|r|
      r.virtual = false
      others = $game.get_robots(r.row, r.col)
      r.virtual = true if others.size > 1
    }
  end

  def prep_next_turn
    $game.dead_robots.each {|r| r.revive}
    toggle_virtual
    $game.record(Action::PAUSE)
    # remove any robots with no lives
    remove = []
    for robot in $game.game_robots
      remove << robot if robot.lives == 0
    end
    remove.each { |r| $game.game_robots.delete(r) }
  end

  def end_game_by_objective
    @log.info "Game Over"
    #shout $game.gametype.text($game)
    $game.end_game
  end

end
