# logic and init

class Robot
  attr_accessor :name, :clientid, :facing, :row, :col, :prevpos
  attr_accessor :maxdamage, :lives, :dead, :spawn, :registers
  attr_accessor :program_complete, :hand, :old_hand
  attr_accessor :image, :image_virtual
  attr_reader :damage, :virtual

  def initialize(clientid, name)
    @clientid = clientid
    @name = name
    defaults
  end

  # when client gets a copy, don't make a new robot
  def set_state(r)
    @name = r.name
    @dead = r.dead
    @registers = r.registers
    @damage = r.damage
    @maxdamage = r.maxdamage
    @lives = r.lives
    @facing = r.facing
    @program_complete = r.program_complete
    @hand = r.hand
    @old_hand = nil
  end

  def defaults
    @dead = false
    @registers = Registers.new
    @damage = 0
    @maxdamage = 9
    @lives = 3
    @facing = Facing.n
    # whether the player is ready to start the turn
    @program_complete = false
    @hand = []
  end

  # die! will record the death type
  def die!(deathtype=Action::REXPLODE)
    return if @dead
    @dead = true
    $game.record(deathtype) # record position before resetting position
    @row, @col = -1, -1
    @lives -= 1
  end
  
  # kills a robot if damage >= maxdamage
  def die_if_dead!
    die! if @damage >= @maxdamage 
  end

  def revive
    @row, @col = @spawn
    @damage = 2
    @dead = false
    $game.record(Action::REVIVE)
  end

  def repair_damage(dmg)
    @damage -= dmg
    @damage = 0 if @damage < 0
    @damage
  end

  def take_damage(dmg)
    @damage += dmg
    @damage = @maxdamage if @damage > @maxdamage
    @damage
  end

  def virtual=(bool)
    @virtual = bool
  end

  def virtual?; @virtual end
  def real?; not @virtual end
  def alive?; not @dead end
  def dead?; @dead end

  def getx; get_pos(@col) end
  def gety; get_pos(@row) end

  def execute_phase(num)
    return if @dead
    send(@registers[num].card.name)
  end
    
  # returns affected robot
  def fire_laser
    return if virtual?
    robot = $game.board.robot_in_los(self, @facing)
    return nil if robot.nil?
    robot.take_damage 1 unless robot.virtual?
    robot
  end

  def ==(other)
    return false unless other.kind_of?(Robot)
    @clientid == other.clientid
  end

  def ===(other)
    self == other
  end

  def <=>(other)
    @registers[$game.phase].card <=> other.registers[$game.phase].card 
  end

  # might want register too, so other players can see cards played...
  def copy
    r = clone
    r.facing = @facing.clone
    r
  end

  def get_pos(pos)
    $mainsize * pos
  end

  def to_s
    "<Robot=#{@clientid}:#{@facing.to_s}:dmg=#{@damage}/#{@maxdamage}:#{@row},#{@col}>"
  end
  alias inspect to_s

  ### handle registers and cards

  # on server side, we receive the play and must adjust our hands
  # a play is the set of non-damage cards in submitted hand
  # move to robots
  def handle_play(play)
    validate_play(play)
    # fill up registers with copies
    @registers.assign(play.collect{|c| c.copy})
  end

  # checks for cheating by comparing submitted hand with server's copy
  def validate_play(play)
    play.each {|p|
      raise CheaterError, "bad hand" if (idx = @old_hand.index(p) ).nil?
      @old_hand.delete_at(idx)
    }
    raise CheaterError, "wrong size" if play.size != @registers.unlocked.size
  end

end
