require 'singleton'

require 'bf/definitions'

require 'bf/server/log'
require 'bf/server/recv'
require 'bf/server/logic'
require 'bf/server/recvgame'

require 'bf/state/gamestate'
require 'bf/net/session'

require 'bf/facing'
require 'bf/robots/include'
require 'bf/elements/include'
require 'bf/factory'

require 'bf/games/include'
require 'bf/games/servergame'
require 'bf/games/serverturn'

# main server mehods: connect, communication, threads
class GameServer
  include Singleton
  attr_accessor :serveruri, :incoming, :clients

  def initialize
    @log = Logger['bf::server']
    @log.info("=============== INITIALIZING BF GAMESERVER ===============")
    @mutex = Mutex.new
    @clients = Hash.new
    $game = GameState.new
  end

  def shutdown(reason="unknown reason")
    @log.info "Shutting down server (#{reason})"
    netsend_all(:recv_server_disconnect, reason)
  end
  
  # quit from server side
  def server_exit
    shutdown("server quitting")
    exit
  end

  def synch
    @mutex.synchronize { yield }
  end
 
  def each_client
    @clients.each_value{|c| yield c}
  end

  def each_clientid
    @clients.each_key{|i| yield i}
  end

  # removes a client
  def remove(clientid)
    c = @clients.delete(clientid)
    c.close if c
  end

  # invoke client method with data object
  def netsend(clientid, method, obj=nil) 
    synch do
      trap_neterr(clientid) { @clients[clientid].session_send(method, obj) }
    end
  end

  # netsend to all clients
  def netsend_all(method, obj=nil)
    each_clientid { |i| netsend(i, method, obj) }
  end

  # what to do when netsend chokes
  def trap_neterr(clientid)
    begin yield
    rescue Exception => e
      @log.error "netsend: Bad session trapped for clientid #{clientid}"
      @log.debug {e}
      name = "unknown"
      player = $game.robots[clientid]
      name = $game.robots[clientid].name.dup if player
      remove(clientid)
      $game.remove(clientid)
      netsend_all :recv_cmsg, "#{name} disconnected (bad session)"
      @mutex.unlock
    end
  end

  # wrappers to send text messages
 
  def relay_cmsg(clientid, message)
    name = $game.robots[clientid].name
    mymsg = ">" + name + "< " + message
    theirmsg = "<" + name + "> " + message
    each_clientid do |cid|
      netsend( cid, :recv_cmsg, (cid == clientid ? mymsg : theirmsg) )
    end
    @log.chat theirmsg
  end

  def send_cmsg(clientid, message)
    netsend( clientid, :recv_cmsg, message )
  end

  def shout(message)
    netsend_all( :recv_cmsg, message )
    @log.info message
  end

  ##### basic server stuff: start, client management, etc
 
  def running?; @running end

  def start(uri)
    netstart(uri)
  end

  MAX_CLIENTS = 3

  def netstart(uri)
    @serveruri = uri if uri
    if @serveruri =~ %r{^bf://(.*):(.*)}
      host = ($1 == "" ? nil : $1)
      port = $2.to_i
    else
      raise ArgumentError, "Invalid uri", caller
    end

    @tcpserver = TCPServer.new(host, port)

    # creates the server threads that handles client connections
    MAX_CLIENTS.times do |id|
      id += 1 # kludge to load img_robot#{id}
      Thread.new {
        while true
          client = @tcpserver.accept # waits for a client connection
          # ok, we have a connection, so get the show on the road
          session = Session.new(client)
          @clients.store(id, session)
          if session.connected?
            begin
              while (msg = session.session_get)
                if msg[1].nil? then self.send(msg[0], id)
                else self.send(msg[0], id, msg[1])
                end
              end
            rescue EOFError, IOError => e
              #@log.warn "Client terminated abruptly."
              #@log.debug {e}
              name = "unknown"
              player = $game.robots[id]
              name = $game.robots[id].name.dup if player
              remove(id)
              $game.remove(id)
              netsend_all :recv_cmsg, "#{name} disconnected (unknown)"
            rescue Exception => e2
              @log.debug {e2}
            end
          end
        end
      }
    end

    @running = true
  end
end
