class EventListener
  attr_reader :bindings
  @@glob_bind = nil
  def initialize
    @bindings = Hash.new
    @@glob_bind = Hash.new unless @@glob_bind
  end
  # connect methods and blocks to signals, only one method per signal
  def connect(sig, &a_proc)
    @bindings[sig] = a_proc
  end
  def connect_method(sig, methodname)
    @bindings[sig] = method(methodname)
  end
  def connect_glob_method(sig, methodname)
    @@glob_bind[sig] = [] if not @@glob_bind[sig]
    @@glob_bind[sig] << method(methodname)
  end
  def glob_bindings
    @@glob_bind
  end
end

# widget should be sortable by what's on top
class Widget < EventListener
  @@nextid = 0
  include GuiTools
  attr_accessor :x, :y, :w, :h, :label, :children, :parent
  attr_accessor :surface, :resized
  attr_reader :realize, :dirty, :widgetid

  @@focused_widget = nil
  @@key_focused_widget = nil
  @@root_surface = nil

  # x and y are relative to the parent widget, so they describe a rect in
  # the parent's coordinate spcae
  # the default is to realize and draw
  def initialize(hash={})
    super()
    @x, @y, @w, @h = ifrom_hash(hash, :x, :y, :w, :h)
    @label = (hash[:label] or "")
    @children = []
    @dirty = true
    @realized = true
    @resized = false
    @widgetid = @@nextid
    @@nextid += 1
  end

  # to initialize sdl stuff
  def init_sdl
  end

  # recursively calls init_sdl, called from MainWindow
  def init_all
    init_sdl
    for child in @children
      child.init_all
    end
    resize if resized?
  end

  def add(*kids)
    kids.each {|child| child.parent = self}
    @children += kids
  end

  def remove(kid)
    k = @children.delete(kid)
    k.parent = nil if k
    k
  end

  def ==(other)
    @widgetid == other.widgetid
  end

  def ===(other)
    self == other
  end

  # when a widget needs to steal the attention of the event manager
  def focus
    @@focused_widget = self
  end
  def unfocus
    @@focused_widget = nil if @@focused_widget and @@focused_widget == self
  end
  # when a key focus is needed
  def key_focus
    @@key_focused_widget = self
  end
  def key_unfocus
    @@key_focused_widget = nil
  end

  def resized?; @resized end
  def dirty?; @dirty end
  def realized?; @realized end
  def clean
    @dirty = false
  end
  def dirty
    @dirty = true 
    @parent.dirty if @parent and not @parent.dirty?
  end
  def realize
    @realized = true
    dirty
  end
  def hide
    @realized = false 
    dirty
  end
  def show; realize end
  def resize
  end

  # transforms a coord point in this child space to parent space
  def child_to_parent(cx, cy)
    [cx + @x, cy + @y]
  end
  # transforms a coord point in parent space to this child space
  def parent_to_child(px, py)
    [px - @parent.x, py - @parent.y]
  end
  # translate root coords to parent coords (recursive)
  def root_to_parent(rx, ry)
    my_x, my_y = @parent.root_to_parent(rx, ry)
    [my_x - @x, my_y - @y]
  end

  # is the parent coord point inside my rect?
  def inside?(px, py)
    return true if px > @x and py > @y and px < (@x + @w) and py < (@y + @h)
    false
  end
  # given root coords
  def root_inside?(rx, ry)
    px, py = @parent.root_to_parent(rx, ry)
    inside?(px, py)
  end

  # finds the widget to pass an event to
  def find_widget(px, py)
    @children.each do |child|
      if child.inside?(px-@x, py-@y)
        return child.find_widget(px-@x, py-@y)
      end
    end
    self
  end

  # draw onto root surface
  def draw(root_surf, dx, dy)
    return unless realized? and dirty?
    root_surf.put(@surface, dx+@x, dy+@y)
    @children.each do |child|
      child.draw(root_surf, dx+@x, dy+@y)
    end
    clean
  end

  # convenience
  def create_surface(sw, sh)
    SDL::Surface.new(Defs::Gfx::SDLFLAGS, sw, sh, @@root_surface)
  end
end

class RootWidget < Widget
  def initialize
    super(Hash.new)
  end
  def dirty; end
  def root?; true end
  def root_to_parent(rx,ry)
    [rx, ry]
  end
end
